

import re
from sys import argv

def tsplit(string, delimiters):
    """Behaves str.split but supports multiple delimiters."""
    
    delimiters = tuple(delimiters)
    stack = [string,]
    
    for delimiter in delimiters:
        for i, substring in enumerate(stack):
            substack = substring.split(delimiter)
            stack.pop(i)
            for j, _substring in enumerate(substack):
                stack.insert(i+j, _substring)
            
    stack.remove('')
    return stack    


def check_baseline(val_list,number,label):
  if abs(min(val_list)) > abs(max(val_list)):
      # peaks should all be negative
      if max(val_list) < 0.0:
        # these peaks are OK
        correction = 0.0
      else:
        correction = min(val_list)*1.5
        # some weak peaks are positive, need to add a constant ...
        print " flipping: ",number,label
 #      print val_list
  else:
      # peaks should all be positive
      if min(val_list) > 0.0:
        # these peaks are OK
        correction = 0.0
      else:
        correction = min(val_list)*1.5
        # some weak peaks are positive, need to add a constant ...
        print " flipping: ",number,label
 #      print val_list

  return correction


if __name__=="__main__":

  title = """
                                 _____  _ _     _   
                                / __  \| (_)   | |  
            ___ ___ _ __  _ __  `' / /'| |_ ___| |_ 
           / __/ __| '_ \| '_ \   / /  | | / __| __|
          | (_| (__| |_) | | | |./ /___| | \__ \ |_ 
           \___\___| .__/|_| |_|\_____/|_|_|___/\__|
                   | |                              
                   |_|                              


  """
  print title


  if len(argv) < 4:
    print "%s field Tcp file1 file2 ..."%argv[0]
    quit()

  field = argv[1]
  Tcp = float(argv[2])
  files = argv[3:]
  block = {}

# itcp_list = [0.0, 0.0, 100.0, 100.0, 200.0, 200.0, 300.0, 400.0, 500.0, 600.0, 800.0, 1000.0, 1200.0, 1600.0, 2000.0, 2400.0, 3200.0]
  itcp_list = []

# print field
# print line1
# print line2

  # build keys
  line_pass = 0
  line_found = 0
  for file in files:
    print " ... processing file:",file
    fh = open(file)
    for line in fh:
      # deal with crazy ccpn format
      if '/' in line:
        print line[:-1]
        data = line.split()
        # Proton label
        label = data[5]
        labels = tsplit(label,('(','/',')'))
        t_labels = labels[0:2]
        t_labels.sort()
        newlabel = "(%s)(%s)%s"%(t_labels[0],t_labels[1],labels[2])
        newline = line.replace(label,newlabel)
        line = newline
        # Nitrogen label
        label = data[6]
        labels = tsplit(label,('(','/',')'))
        t_labels = labels[0:2]
        t_labels.sort()
        newlabel = "(%s)(%s)%s"%(t_labels[0],t_labels[1],labels[2])
        newline = line.replace(label,newlabel)
        line = newline
        print line[:-1]
      data = line.split()

      if "Height" in line:
        # header
#       print line      
        line_pass += 1
        pass
      elif "NOT SET" in line:
        # header
#       print line      
        line_pass += 1
        pass
      else:
        line_found += 1
        label = data[5].strip()
        label = label.split("]")[0]
        label = label.split("[")[-1] 
        label = label.strip()
        itcp = float(data[4])
        height = (float(data[8]))
        volume = (float(data[9]))
#       print label, itcp, height, volume
        if label in block:
          block[label].append([itcp,height,volume])
        else:
          block[label] = []
          block[label].append([itcp,height,volume])
    fh.close()
  # END OF DATA READING
  print " ... lines found :",line_found
  print " ... lines passed:",line_pass
  print " ... total lines :",line_pass+line_found
  

  fh_int = open("%s_int.list"%field,"w")
  fh_vol = open("%s_vol.list"%field,"w")
  counter = 0
  for i in block:
    p = block[i]
    p.sort()
    parts = re.split('(\D.*)',i)

    # generate number and label
    if len(parts) > 2:
      if parts[0] == '':
        number = '0' 
      else:
        number = parts[0]
      label = parts[1]
    else:
      number = parts[0]
      label = "unass"
    print number, label

    counter += 1
    check_list = []
    int_list = []
    vol_list = []
    for k in p:
      check_list.append(k[0])
      int_list.append(k[1])
      vol_list.append(k[2])

    # check the baselines don't become negative
    int_correction = check_baseline(int_list,number,label)
    vol_correction = check_baseline(vol_list,number,label)

    int_line = "%i %s %s"%(counter, number, label ) 
    vol_line = "%i %s %s"%(counter, number, label ) 
    for k in range(len(int_list)):
      int_line = "%s %f"%(int_line,abs(int_list[k]-int_correction))
      vol_line = "%s %f"%(vol_line,abs(vol_list[k]-vol_correction))


    # start some checking      
    strict = False 
    if (("/" in label) and (strict == True)):
      print " discarded:",number,label
    elif "{*}" in label:
      print " discarded:",number,label
    else:
      if itcp_list == []:
         # generate header lines
         itcp_list = check_list
         line1 = ""
         line2 = ""
         for i in enumerate(itcp_list):
           if i[1] == 0.0:
             line1 = ("%s%.4f "%(line1,0.00))
             line2 = ("%s%.1f "%(line2,i[1]))
           else:
             line1 = ("%s%.4f "%(line1,Tcp))
             line2 = ("%s%.1f "%(line2,i[1]))
         # write out headers
         fh_int.write("%s\n"%field)
         fh_int.write("%s\n"%line1)
         fh_int.write("%s\n"%line2)
         fh_vol.write("%s\n"%field)
         fh_vol.write("%s\n"%line1)
         fh_vol.write("%s\n"%line2)
         # write out first line
         fh_int.write("%s\n"%int_line)
         fh_vol.write("%s\n"%vol_line)

         print " # points :",len(itcp_list)
      elif check_list == itcp_list:
        # write out data lines
        fh_int.write("%s\n"%int_line)
        fh_vol.write("%s\n"%vol_line)
      else:
        print "error with set:",label, number
        print "error check_list:",check_list
        print "error itcp_list:",itcp_list

