# Script to handle clustering of residues in global fits of CPMG data
# Jonathan C. Lansing	11/01/2004

use strict;
require		Exporter;		# Use the Exporter package

# Statistics package that includes F distribution calculations
use Statistics::Distributions;


# Simple object to keep track of the members of each cluster {MEMBER}
# and the cluster to which a given set belongs {CLUSTER}
package Glove::Member;
my @ISA		= qw(Exporter Autoloader);

sub clear {
	my $self = shift;
	delete($self->{CLUSTER});
	delete($self->{MEMBERS});
}

sub set_cluster {
	my $self = shift;
	my $s = shift;
	my $c = shift;
	my $old = $self->{CLUSTER}->{$s};
	if (defined($old)) { $self->remove_member($old,$s); }
	$self->{CLUSTER}->{$s} = $c;
	push(@{$self->{MEMBERS}->{$c}},$s);
}

sub get_cluster {
	my $self = shift;
	my $s = shift;
	return $self->{CLUSTER}->{$s};
}

sub remove_member {
	my $self = shift;
	my $c = shift;
	my $s = shift;
	my $i;
	for ($i = 0; $i <= $#{$self->{MEMBERS}->{$c}}; $i++) {
		if ($self->{MEMBERS}->{$c}->[$i] eq $s) {
			splice(@{$self->{MEMBERS}->{$c}},$i,1);
			last;
		}
	}
}

sub set_members {
	my $self = shift;
	my $c = shift;
	my $s;
	my $old;
	while ($s = shift) {
		$old = $self->{CLUSTER}->{$s};
		if (defined($old)) { $self->remove_member($old,$s); }
		$self->{CLUSTER}->{$s} = $c;
		push(@{$self->{MEMBERS}->{$c}},$s);
	}
}

sub get_members {
	my $self = shift;
	my $c = shift;
	if (defined($self->{MEMBERS}->{$c})) { 
		return sort @{$self->{MEMBERS}->{$c}};
	} else { return; }
}

sub get_all {
	my $self = shift;
	my @list;
	my $c;
	foreach $c (keys(%{$self->{MEMBERS}})) {
		push(@list,@{$self->{MEMBERS}->{$c}});
	}
	return @list;
}



################################################################################

package Glove::Cluster;
@ISA		= qw(Exporter Autoloader);

# Object to handle clustering decisions





# Duplicate almost anything
# The targets of references are DUPLICATED -- the new reference will point
# to the copy, instead of producing another pointer to the same data
sub duplicate {
        my $a = shift;
        my $i;
        my $b;
        if (!defined($a)) { return $a; }
        my $type = ref($a);
        if (!ref($a)) {
                # Not a reference
                $b = $a;
        } elsif ($type eq "SCALAR") {
                # It's a scalar reference
                $$b = $$a;
        } elsif ($type eq "ARRAY") {
                # It's an array reference
                foreach $i (@$a) {
                        $b->[$i] = duplicate($a->[$i]);
                }
        } elsif ($type eq "CODE") {
                print STDERR "ERROR: copyit() can't handle type \'CODE\'\n";
                $b = $a;
        } elsif ($type eq "GLOB") {
                print STDERR "ERROR: copyit() can't handle type \'GLOB\'\n";
                $b = $a;
        } elsif ($type eq "REF") {
                # It's a reference to a reference
                $$b = duplicate($$a);
        } elsif ($type eq "HASH") {
                # It's a hash
                foreach $i (keys(%$a)) {
                        $b->{$i} = duplicate($a->{$i});
                }
        } else {
                # It's an object reference
		$b = {};
		bless($b,$type);
                foreach $i (keys(%$a)) {
                        $b->{$i} = duplicate($a->{$i});
                }
        }
        return $b;
}

sub read {
	my $self = shift;
	my $a = shift;
	my @list = @_;
	my $p;
	my $i;
	my $c = $self->get_active();
	if (ref($a) eq "Glove::Output") {
		if ($c eq "???" || $c eq "OUT" || $c eq "TEST") {
			# Don't import anything
		} else {
			# Import the global values
			my @list = $self->list_current();
			my $i = $list[0];
			foreach $p (keys(%{$self->{GBL}})) {
				$self->{GPAR}->{$c}->{$p}  = $a->get_par($i,$p);
				$self->{DGPAR}->{$c}->{$p} = $a->get_dpar($i,$p);
			}
		}
		$self->{DF} = duplicate($a->get_df());
		$self->{X2} = duplicate($a->get_x2());
	} elsif (ref($a) eq "Glove::Cluster") {
		$self->{RX2} = duplicate($a->get_x2());
		$self->{RDF} = duplicate($a->get_df());
	} else {
		die "ERROR (Glove::Cluster::read()): No object provided\n";
	}
	return;
}

sub reference {
	my $self = shift;
	my $a = shift;
	if (ref($a) eq "Glove::Output") {
		$self->{RDF} = duplicate($a->get_df());
		$self->{RX2} = duplicate($a->get_x2());
		$self->{PAR} = duplicate($a->get_par());
		$self->{DPAR} = duplicate($a->get_dpar());
	} else {
		die "ERROR (Glove::Cluster::reference()): No object provided\n";
	}
	return;
}

sub set_gbl {
	my $self = shift;
	my $p = shift;
	my $v = shift;
	my $sf = shift;
	if (!defined($p)) { die "ERROR: Glove::Cluster::set_gbl() requires a parameter name\n"; }
	$self->{GBL}->{$p} = $v;
	$self->{GSF}->{$p} = $sf;

}

sub list_gbl {
	my $self = shift;
	my $null;
	if (!defined($self->{STAGE}) || $self->{STAGE} eq "NEW") {
		return;
	} else {
		return keys(%{$self->{GBL}});
	}
}

sub get_gbl {
	my $self = shift;
	my $p = shift;
	if (!defined($p)) { die "ERROR: Glove::Cluster::get_gbl() requires a parameter name\n"; }
	return $self->{GBL}->{$p};
}

sub gbl_exists {
	my $self = shift;
	my $p = shift;
	if (!defined($p)) { die "ERROR: Glove::Cluster::gbl_exists() requires a parameter name\n"; }
	return exists($self->{GBL}->{$p});
}

sub get_x2 {
	my $self  = shift;
	my $set   = shift;
	if (!defined($set)) { return $self->{X2}; }
	return $self->{X2}->{$set};
}

sub get_df {
	my $self  = shift;
	my $set   = shift;
	if (!defined($set)) { return $self->{DF}; }
	return $self->{DF}->{$set};
}

sub get_par {
	my $self = shift;
	my $c = shift;
	my $p = shift;
	if (defined($p)) { return $self->{GPAR}->{$c}->{$p}; }
	if (defined($c)) { return $self->{GPAR}->{$c}; }
	else { return $self->{GPAR}; }
}

sub get_dpar {
	my $self = shift;
	my $c = shift;
	my $p = shift;
	if (defined($p)) { return $self->{DGPAR}->{$c}->{$p}; }
	if (defined($c)) { return $self->{DGPAR}->{$c}; }
	else { return $self->{DGPAR}; }
}

sub get_ratio {
	my $self = shift;
	my $s = shift;
	if (defined($s)) { return $self->{RATIO}->{$s}; }
	else { return $self->{RATIO}; }
}

sub get_f {
	my $self = shift;
	my $s = shift;
	if (defined($s)) { return $self->{F}->{$s}; }
	else { return $self->{F}; }
}

sub get_choice {
	my $self = shift;
	my $s = shift;
	if (defined($s)) { return $self->{DECISION}->{$s}; }
	else { return $self->{DECISION}; }
}

sub get_good {
	my $self = shift;
	return $self->{GOOD};
}

sub get_alpha {
	my $self = shift;
	return $self->{ALPHA};
}

sub compare {
	my $self = shift;
	my $s = shift;
	my $a = $self->{$self->{STAGE}};
	$self->{NACCEPT} = $self->{NREJECT} = 0;
	$self->{ACCEPT} = [];
	$self->{REJECT} = [];
	foreach $s (keys(%{$self->{X2}})) {
		if (!defined($self->{RX2}->{$s})) {
			die "ERROR (Glove::Cluster::compare()): Reference X2 not defined for set \'$s\'\n";
		}
		if (!defined($self->{RDF}->{$s})) {
			die "ERROR (Glove::Cluster::compare()): Reference dfree not defined for set \'$s\'\n";
		}
		if (!defined($self->{DF}->{$s})) {
			die "ERROR (Glove::Cluster::compare()): Dfree not defined for set \'$s\'\n";
		}
		$self->{F}->{$s} = Statistics::Distributions::fdistr(
			$self->{RDF}->{$s},$self->{DF}->{$s},$a);
		$self->{RATIO}->{$s} = $self->{X2}->{$s} / $self->{RX2}->{$s};
		if ($self->{RATIO}->{$s} < $self->{F}->{$s}) {
			push(@{$self->{ACCEPT}},$s);
			$self->{NACCEPT}++;
		} else {
			push(@{$self->{REJECT}},$s);
			$self->{NREJECT}++;
		}
	}
	return $self->{NACCEPT};
}


sub set_conf_expand {
	my $self = shift;
	my $a = shift;
	if (!defined($a)) { die "ERROR: Glove::Cluster::set_conf_expand() requires a value\n"; }
	$self->{EXPAND} = $a;
}

sub set_conf_contract {
	my $self = shift;
	my $a = shift;
	if (!defined($a)) { die "ERROR: Glove::Cluster::set_conf_contract() requires a value\n"; }
	$self->{CONTRACT} = $a;
}

sub set_tol {
	my $self = shift;
	my $p = shift;
	my $tol = shift;
	if (!defined($tol)) { die "ERROR: Glove::Cluster::set_tol() requires a parameter name & value\n"; }
	if (!defined($self->{TOL})) { $self->{TOL} = {}; }
	$self->{TOL}->{$p} = $tol;
}

sub get_ntol {
	my $self = shift;
	my @list = keys(%{$self->{TOL}});
	return $#list + 1;
}

sub set_bad {
	my $self = shift;
	my $s = shift;
	if (!defined($s)) { die "ERROR: Glove::Cluster::set_bad() requires a set name\n"; }
	$self->{CLUSTER}->set_cluster($s,"BAD");
}

sub set_names {
	my $self = shift;
	my $i;
	$self->{SETS} = @_;
	if (!defined($self->{CLUSTER})) { $self->{CLUSTER} = bless({},"Glove::Member"); }
	$self->{CLUSTER}->clear();
	foreach $i (@_) {
		$self->{CLUSTER}->set_cluster($i,"???");
	}
}

sub list_cluster {
	my $self = shift;
	my $c = shift;
	if (!defined($c)) { die "ERROR: Glove::Cluster::list_cluster() requires a cluster name\n"; }
	return $self->{CLUSTER}->get_members($c);
}

sub size_cluster {
	my $self = shift;
	my $c = shift;
	my @list;
	if (!defined($c)) { die "ERROR: Glove::Cluster::list_cluster() requires a cluster name\n"; }
	@list = $self->{CLUSTER}->get_members($c);
	return $#list;
}

sub current {
	my $self = shift;
	if (!defined($self->{CURRENT})) { $self->{CURRENT} = "???"; }
	return $self->{CURRENT};
}

sub list_current {
	my $self = shift;
	if (!defined($self->{CURRENT})) { $self->{CURRENT} = "???"; }
	my $c = $self->{CURRENT};
	if (defined($self->{STAGE}) && $self->{STAGE} eq "EXPAND") {
		$c = "???";
	}
	return $self->{CLUSTER}->get_members($c);
}

sub fixed {
	my $self = shift;
	my $p = shift;
	my $null;
	if (defined($self->{STAGE}) && $self->{STAGE} eq "EXPAND") {
		return $self->{GPAR}->{$self->{CURRENT}}->{$p};
	} else {
		return;
	}
}

sub clust_name {
	my $self = shift;
	my $s = shift;
	if (!defined($s)) { die "ERROR: Glove::Input::clust_name() requires a set name\n"; }
	return $self->{CLUSTER}->get_cluster($s);
}

sub is_current {
	my $self = shift;
	my $s = shift;
	my $a;
	if ($self->{CLUSTER}->get_cluster($s) eq $self->{CURRENT}) { return 1; }
	else { return 0; }
}

sub is_active {
	my $self = shift;
	my $s = shift;
	my $a;
	my $c = $self->{CURRENT};
	if ($self->{STAGE} eq "EXPAND") { $c = "???"; }
	if ($self->{CLUSTER}->get_cluster($s) eq $c) { return 1; }
	else { return 0; }
}

sub get_active {
	my $self = shift;
	if (!defined($self->{STAGE}) || $self->{STAGE} eq "EXPAND") { 
		return "???"; 
	} else { return $self->{CURRENT}; }
}


sub filter_x2 {
	my $self = shift;
	my $crit = shift;
	my @list;
	my $s;
	my $c;
	if (!defined($crit)) {
		die "ERROR: Glove::Cluster::filter_x2() called without a cutoff value\n";
	}
	foreach $s ($self->{CLUSTER}->get_members("???")) {
		if (defined($self->{RX2}->{$s}) && $self->{RX2}->{$s}  > $crit) {
			$self->set_bad($s);
			push(@list,$s);
		}
	}
	if (@list) {
		print "Removed sets with X2i > $crit (@list)\n";
	}
}

sub report_choice {
	my $self = shift;
	my $good = shift;
	my $bad = shift;
	my @list = $self->{CLUSTER}->get_all();
	my $p;
	my $s;
	my $c = $self->{CURRENT};
	if (defined($bad)) {
		print "Current_cluster\t$c\n";
		print "GPAR\tVAL\n";
		foreach $p (keys(%{$self->{GBL}})) {
			printf "$p\t%5.3f",$self->{GPAR}->{$c}->{$p};
			printf "\t(%5.3f)",$self->{DGPAR}->{$c}->{$p};
			print "\n";
		}
		print "Set\tClust\tX2[0]\tdf[0]\tX2\tdf\tRatio\tF\tDecision\n";
	} else {
		print "Inital fits (independent)\n";
		print "Set\tClust\tX2[0]\tdf[0]";
		foreach $p ($self->list_gbl()) { print "\t$p"; }
		print "\n";
	}
	foreach $s (sort @list) {
		print "$s\t";
		print $self->clust_name($s);
		printf "\t%5.3f",$self->{RX2}->{$s};
		printf "\t%u",$self->{RDF}->{$s};
		if (defined($bad) && $self->is_active($s)) {
			printf "\t%5.3f\t",$self->{X2}->{$s};
			printf "%u\t",$self->{DF}->{$s};
			printf "%5.3f\t",$self->{RATIO}->{$s};
			printf "%5.3f\t",$self->{F}->{$s};
			if ($self->{DECISION}->{$s}) { print "$good"; }
			else { print "$bad"; }
		} elsif (!defined($bad)) {
			foreach $p ($self->list_gbl()) {
				printf "\t%5.3f",$self->{PAR}->{$s}->{$p};
			}
		}
		print "\n";
	}
	print "\n";
}

# Decide what to do with the last iteration's info
sub refine {
	my $self = shift;
	my $good;
	my $bad;
	my $s;
	my @list;
	if (!defined($self->{STAGE})) { 
		$self->{STAGE} = "PICK";
		$self->{CURRENT} = -1;
		$self->report_choice();
	} elsif ($self->{STAGE} eq "EXPAND") {
		# Cluster expansion
		$good = "ADD";
		$bad = "OMIT";
		$self->compare($self->list_current());
		$self->report_choice($good,$bad);
		if ($self->{NACCEPT} == 0) {
			# Cluster size is stable.  Pick a new one
			$self->{STAGE} = "PICK";
		} else {
			$self->{STAGE} = "CONTRACT";
			# Check if this test has been done already
			if ($self->find_pick(@{$self->{ACCEPT}}),
				$self->list_current()) {
				# Already tried this pick -- no need for redundant calc
				print "# Expanded cluster already tried.";
				print "# Cluster $self->{CURRENT} is done.\n";
				$self->{STAGE} = "PICK";
			} else {
				# Add the good ones to the cluster
				$self->{CLUSTER}->set_members($self->{CURRENT},
					@{$self->{ACCEPT}});
				$self->{STAGE} = "CONTRACT";
				$self->add_pick($self->{CLUSTER}->
					get_members($self->{CURRENT}));
			}
		}
	} elsif ($self->{STAGE} eq "CONTRACT") {
		# Cluster contraction
		$good = "KEEP";
		$bad = "OMIT";
		$self->compare($self->list_current());
		$self->report_choice($good,$bad);
		if ($self->{NREJECT} == 0) {
			# Cluster size is stable.
			if ($self->size_cluster("???") < 0) {
				# All residues assigned
				$self->{STAGE} = "DONE";
			} else {
				# Still some unassigned residues.  Try to expand it.
				$self->{STAGE} = "EXPAND";
			}
		} elsif ($self->{NACCEPT} < 2) {
			# Cluster has vanished.
			print "# Cluster $self->{CURRENT} has vanished\n";
			$self->{CURRENT} -= 1;
			$self->{STAGE} = "PICK";
		} else {
			# Trim down the cluster
			foreach $s (@{$self->{REJECT}}) {
				$self->{CLUSTER}->set_cluster($s,"???");
			}
			$self->add_pick($self->{CLUSTER}->get_members($self->{CURRENT}));
			$self->{STAGE} = "CONTRACT";
		}
	}
	if ($self->{STAGE} eq "PICK") {
		# Pick a new cluster
		$self->pick_cluster();
	}
	if ($self->{STAGE} eq "DONE") {
		print "** Clustering complete **\n";
	} elsif ($self->{STAGE} eq "CONTRACT") {
		@list = $self->{CLUSTER}->get_members($self->get_active());
		print "Next step: optimize cluster $self->{CURRENT} (@list)\n";
	} elsif ($self->{STAGE} eq "EXPAND") {
		@list = $self->{CLUSTER}->get_members($self->get_active());
		print "Next step: expand cluster $self->{CURRENT} to include (@list)\n";
	}
	print "\n";
	return $self->{STAGE};
}

sub pick_cluster {
	my $self = shift;
	my @list = $self->{CLUSTER}->get_members("???");
	my $x;
	my $y;
	my $cl = {};
	my $flag;
	my $p;
	foreach $x (@list) {
		push(@{$cl->{$x}},$x); 
		foreach $y (@list) {
			if ($x eq $y) { next; }
			$flag = 1;
			foreach $p (keys(%{$self->{TOL}})) {
				if ((2 * $self->{TOL}->{$p} * 
					($self->{PAR}->{$x}->{$p} 
					- $self->{PAR}->{$y}->{$p})
					/ ($self->{DPAR}->{$x}->{$p} 
					+ $self->{DPAR}->{$y}->{$p}))**2 
					> 1.0) { $flag = 0; }

			}
			if ($flag) { 
				push(@{$cl->{$x}},$y); 
			}
		}
		if ($#{$cl->{$x}} < 1) {
			delete($cl->{$x});
		}
	}
	foreach $x (keys(%$cl)) {
		# Remove picks already tried
		if ($self->find_pick(@{$cl->{$x}})) { delete($cl->{$x}); }
	}
	if (@list = keys(%$cl)) {
		# Find the largest cluster
		@list = sort { $#{$cl->{$b}} <=> $#{$cl->{$a}} } keys(%$cl);
		$x = $list[0];
		@list = @{$cl->{$x}};
		$self->{CURRENT} += 1;
		$self->{CLUSTER}->set_members($self->{CURRENT},@list);
		$self->add_pick(@list);
		$self->{STAGE} = "CONTRACT";
	}
	else { $self->{STAGE} = "DONE"; }
}




sub add_pick {
	my $self = shift;
	my @list = sort {$a cmp $b} @_;
	my $str = shift(@list);
	my $s;
	while ($s = shift(@list)) { $str .= "_" . $s; }
	push(@{$self->{PICK}},$str);
}

sub find_pick {
	my $self = shift;
	my @list = sort {$a cmp $b} @_;
	my $str = shift(@list);
	my $s;
	while ($s = shift(@list)) { $str .= "_" . $s; }
	foreach $s (@{$self->{PICK}}) {
		if ($s eq $str) { return 1; }
	}
	return 0;
}

sub guess_range {
	my $self = shift;
	my $p = shift;
	if  (!defined($p)) { return; }
	my $c = $self->get_active();
	my $sz = $self->{GBL}->{$p};
	my $sf = $self->{GSF}->{$p};
	my $max;
	my $min;
	my $s;
	if ($c eq "???" || !defined($sf) || !defined($sz)) {
		return;
	} else {
		if (defined($self->{DGPAR}->{$c}->{$p})) {
			# Cluster has already been tested once
			$min = $self->{DPAR}->{$c}->{$p} 
				- $self->{DGPAR}->{$c}->{$p} * $sf;
			$max = $self->{DPAR}->{$c}->{$p} 
				+ $self->{DGPAR}->{$c}->{$p} * $sf;
		} else {
			# Cluster has not been tested
			my $sum = 0;
			my $dsum = 0;
			my $n = 0;
			foreach $s ($self->{CLUSTER}->get_members($c)) {
				$sum += $self->{PAR}->{$s}->{$p};
				$dsum += $self->{DPAR}->{$s}->{$p};
				++$n;
			}
			$sum /= $n;
			$dsum *= $sf / $n;
			$min = $sum - $dsum;
			$max = $sum + $dsum;
		}
		$s = sprintf "%5.5f %5.5f %u",$min,$max,$sz;
		return $s;
	}
}

__END__


Data structure

{EXPAND}		Alpha value for expansion
{CONTRACT}		Alpha value for contraction
{CURRENT}		Current cluster name
{CLUSTER}		Member object (cluster by set name & set list by cluster name)
{X2}->{$s}		Reduced X2 for set $s
{DF}->{$s}		Degrees of freedom for set $s
{RX2}->{$s}		Reference reduced X2 for set $s
{RDF}->{$s}		Reference degrees of freedom for set $s
{GPAR}->{$c}->{$p}	Global parameter $p value for cluster $c
{DGPAR}->{$c}->{$p}	Global parameter $p stdev for cluster $c
{F}->{$s}		Last used F value for set $s
{RATIO}->{$s}		Last used ratio for set $s
{DECISION}->{$s}	Last decision made for set $s
@{ACCEPT}		List of sets accepted in last step
@{REJECT}		List of sets rejected in last step
{NACCEPT}		Number of accept decisions made in last step
{NREJECT}		Number of reject decisions made in last step
{STAGE}			Current stage of clustering

