# Module to parse the output of GLOVE
# Jonathan C. Lansing	02/25/2005

use strict;
require		Exporter;		# Use the Exporter package

package 	Glove::Output;
my @ISA		= qw(Exporter Autoloader);
my @EXPORT_OK	= qw(new file read get_par get_dpar get_x2 get_simval 
			get_simpar get_simdpar get_simx2);

# Set or get the file name
sub file {
	my $self = shift;
	return $self->{FILE};
}

# Read in the data
sub read {
	my $self = shift;
	if (@_) { $self->{FILE} = shift; }
	my $file = $self->{FILE};
	open(my $in, "<" . $file) || die "ERROR: Unable to open file \'$file\'\n";;
	my $stage = 0;
	my $line;
	my $s;		# Set name
	my @e;
	my @label;
	# Wipe any existing values
	undef($self->{X2});
	undef($self->{DF});
	undef($self->{PAR});
	undef($self->{DPAR});
	undef($self->{SIMX2});
	undef($self->{SIMVAL});
	undef($self->{SIMPAR});
	undef($self->{SIMDPAR});
	while ($line = <$in>) {
		chomp($line);
		if ($line =~ /^\s*$/) {
			# Blank line
		} elsif ($stage == 0) {
			# Preliminary stage of the output
			if ($line =~ /^#\sFitting\sresults$/) { ++$stage; }
		} elsif ($stage == 1) {
			# Fit results section
			if ($line =~ /^#\s(Monte\sCarlo|Jackknife)\ssimulations\s\((\d+)\)$/) {
				++$stage;
				$self->{SIMTYPE} = $1;
			} elsif ($line =~ /^SET\s+(\S+)$/) {
				$s = $1;
			} elsif ($line =~ /^\s+X2red\s(\S+)$/) {
				$self->{X2}->{$s} = $1;
			} elsif ($line =~ /^\s+Dfree\s(\S+)$/) {
				$self->{DF}->{$s} = $1;
			} elsif ($line =~ /^\s+(\S+)\s+(\S+)\s+\((\S+)\)/) {
				$self->{PAR}->{$s}->{$1} = $2;
				$self->{DPAR}->{$s}->{$1} = $3;
			} elsif ($line =~ /^\s+(\S+)\s+(\S+)$/) {
				$self->{PAR}->{$s}->{$1} = $2;
				$self->{DPAR}->{$s}->{$1} = 0;
			} elsif ($line =~ /^\s+Global\sX2red\s=\s+(\S+)/) {
				$self->{X2}->{GLOBAL} = $1;
			}
		} elsif ($stage == 2) {
			# Simulation section
			if ($line =~ /^#(Monte\sCarlo|Jackknife)\serror\sanalysis\sresults$/) {
				++$stage;
				$self->{SIM_TYPE} = $1;
			} elsif ($line =~ /^(\d+)\s+(\S+)\s+X2\s+(\S+)/) {
				$self->{SIMX2_IT}->{$2}->[$1] = $3;
			} elsif ($line =~ /^(\d+)\s+(\S+)\s+(\S+)\s+(\S+)/) {
				$self->{SIMVAL}->{$2}->{$3}->[$1] = $4;
			}
		} elsif ($stage == 3) {
			# Simulation results section
			if ($line =~ /^SET\s+(\S+)$/) {
				$s = $1;
			} elsif ($line =~ /^\s+X2red\s(\S+)$/) {
				$self->{SIMX2}->{$s} = $1;
				$self->{SIMDX2}->{$s} = 0;
			} elsif ($line =~ /^\s+X2red\s+(\S+)\s+\((\S+)\)$/) {
				$self->{SIMX2}->{$s} = $1;
				$self->{SIMDX2}->{$s} = $2;
			} elsif ($line =~ /^Global\s+X2red\s+=\s+(\S+)\s+\((\S+)\)$/) {
				$self->{SIMX2}->{GLOBAL} = $1;
				$self->{SIMDX2}->{GLOBAL} = $2;
			} elsif ($line =~ /^\s+Dfree\s(\S+)$/) {
				# Ignore the degrees of freedom ($1) for now
			} elsif ($line =~ /^\s+(\S+)\s+(\S+)\s+\((\S+)\)$/) {
				$self->{SIMPAR}->{$s}->{$1} = $2;
				$self->{SIMDPAR}->{$s}->{$1} = $3;
			}
		}
	}
	close($in);
}

sub get_par {
	my $self  = shift;
	my $set   = shift;
	my $pname = shift;
	if (!defined($set)) { return $self->{PAR}; }
	if (!defined($pname)) { return $self->{PAR}->{$set}; }
	return $self->{PAR}->{$set}->{$pname};
}

sub get_dpar {
	my $self  = shift;
	my $set   = shift;
	my $pname = shift;
	if (!defined($set)) { return $self->{DPAR}; }
	if (!defined($pname)) { return $self->{DPAR}->{$set}; }
	return $self->{DPAR}->{$set}->{$pname};
}

sub get_x2 {
	my $self  = shift;
	my $set   = shift;
	if (!defined($set)) { return $self->{X2}; }
	return $self->{X2}->{$set};
}

sub get_global_x2 {
	my $self  = shift;
	return $self->{X2}->{GLOBAL};
}

sub list_pars {
	my $self  = shift;
	my $set   = shift;
	my @temp;
	if (defined($set)) { return (sort(keys(%{$self->{PAR}->{$set}}))); }
	elsif (@temp = keys(%{$self->{PAR}})) {
		return (sort(keys(%{$self->{PAR}->{$temp[0]}})));
	} else { return; }
}

sub get_simpar {
	my $self  = shift;
	my $set   = shift;
	my $pname = shift;
	if (!defined($set)) { return $self->{SIMPAR}; }
	if (!defined($pname)) { return $self->{SIMPAR}->{$set}; }
	return $self->{SIMPAR}->{$set}->{$pname};
}

sub get_simdpar {
	my $self  = shift;
	my $set   = shift;
	my $pname = shift;
	if (!defined($set)) { return $self->{SIMDPAR}; }
	if (!defined($pname)) { return $self->{SIMDPAR}->{$set}; }
	return $self->{SIMDPAR}->{$set}->{$pname};
}

sub get_simx2 {
	my $self  = shift;
	my $set   = shift;
	if (!defined($set)) { return $self->{SIMX2}; }
	return $self->{SIMX2}->{$set};
}

sub get_simval {
	my $self  = shift;
	my $set   = shift;
	my $pname = shift;
	my $it    = shift;
	if (!defined($set)) { return $self->{SIMVAL}; }
	if (!defined($pname)) { return $self->{SIMVAL}->{$set}; }
	if (!defined($it)) { return $self->{SIMVAL}->{$set}->{$pname}; }
	return $self->{SIMVAL}->{$set}->{$pname}->[$it];
}

sub get_simx2val {
	my $self  = shift;
	my $set   = shift;
	my $it    = shift;
	if (!defined($set)) { return $self->{SIMX2_IT}; }
	if (!defined($it)) { return $self->{SIMX2_IT}->{$set}; }
	return $self->{SIMX2_IT}->{$set}->[$it];
}

sub get_df {
	my $self  = shift;
	my $set   = shift;
	if (!defined($set)) { return $self->{DF}; }
	return $self->{DF}->{$set};
}

sub list_sets {
	my $self = shift;
	return keys(%{$self->{DF}});
}

sub get_sim_type {
	my $self = shift;
	return $self->{SIM_TYPE};
}

# Retrive a parameter setting for one data set from one or all monte carlo steps
sub get_mc_par {
	my $self = shift;
	my $s = shift;
	my $p = shift;
	my $it = shift;
	if ($self->{SIMTYPE} !~ /^monte/i) {
		die "ERROR: Glove::Output::get_mc_par() called on data without MC simulations.\n";
	}
	if (!defined($p)) { die "ERROR: Glove::Output::get_mc_par() requires set and parameter names\n"; }
	if (defined($it)) { return $self->{SIMVAL}->{$s}->{$p}->[$it]; }
	else { 
		# Get rid of the undefined value at the start (Glove reports MC starting at 1)
		my @list = @{$self->{SIMVAL}->{$s}->{$p}};
		shift(@list);
		return @list; 
	}
}
